@@ -18,14 +18,20 @@ Copyright (C) 2010 Hiroki Ohtani(liris)
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 """
+from __future__ import print_function
 
 
+import six
 import socket
 
 try:
     import ssl
     from ssl import SSLError
-    from backports.ssl_match_hostname import match_hostname
+    if hasattr(ssl, "match_hostname"):
+        from ssl import match_hostname
+    else:
+        from backports.ssl_match_hostname import match_hostname
+
     HAVE_SSL = True
 except ImportError:
     # dummy class of SSLError for ssl none-support environment.
@@ -34,13 +40,17 @@ except ImportError:
 
     HAVE_SSL = False
 
-from urlparse import urlparse
+from six.moves.urllib.parse import urlparse
+if six.PY3:
+    from base64 import encodebytes as base64encode
+else:
+    from base64 import encodestring as base64encode
+
 import os
 import array
 import struct
 import uuid
 import hashlib
-import base64
 import threading
 import time
 import logging
@@ -203,8 +213,11 @@ def create_connection(url, timeout=None, **options):
     timeout: socket timeout time. This value is integer.
              if you set None for this value, it means "use default_timeout value"
 
-    options: current support option is only "header".
-             if you set header as dict value, the custom HTTP headers are added.
+
+    options: "header" -> custom http header list.
+             "cookie" -> cookie value.
+             "http_proxy_host" - http proxy host name.
+             "http_proxy_port" - http proxy port. If not set, set to 80.
     """
     sockopt = options.get("sockopt", [])
     sslopt = options.get("sslopt", {})
@@ -215,7 +228,7 @@ def create_connection(url, timeout=None, **options):
     return websock
 
 _MAX_INTEGER = (1 << 32) -1
-_AVAILABLE_KEY_CHARS = range(0x21, 0x2f + 1) + range(0x3a, 0x7e + 1)
+_AVAILABLE_KEY_CHARS = list(range(0x21, 0x2f + 1)) + list(range(0x3a, 0x7e + 1))
 _MAX_CHAR_BYTE = (1<<8) -1
 
 # ref. Websocket gets an update, and it breaks stuff.
@@ -224,7 +237,7 @@ _MAX_CHAR_BYTE = (1<<8) -1
 
 def _create_sec_websocket_key():
     uid = uuid.uuid4()
-    return base64.encodestring(uid.bytes).strip()
+    return base64encode(uid.bytes).decode('utf-8').strip()
 
 
 _HEADERS_TO_CHECK = {
@@ -300,7 +313,7 @@ class ABNF(object):
 
         fin: fin flag. if set to 0, create continue fragmentation.
         """
-        if opcode == ABNF.OPCODE_TEXT and isinstance(data, unicode):
+        if opcode == ABNF.OPCODE_TEXT and isinstance(data, six.text_type):
             data = data.encode("utf-8")
         # mask must be set if send data from client
         return ABNF(fin, 0, 0, 0, opcode, 1, data)
@@ -329,6 +342,8 @@ class ABNF(object):
             frame_header += chr(self.mask << 7 | 0x7f)
             frame_header += struct.pack("!Q", length)
 
+        frame_header = six.b(frame_header)
+
         if not self.mask:
             return frame_header + self.data
         else:
@@ -337,7 +352,11 @@ class ABNF(object):
 
     def _get_masked(self, mask_key):
         s = ABNF.mask(mask_key, self.data)
-        return mask_key + "".join(s)
+
+        if isinstance(mask_key, six.text_type):
+            mask_key = mask_key.encode('utf-8')
+
+        return mask_key + s
 
     @staticmethod
     def mask(mask_key, data):
@@ -348,11 +367,22 @@ class ABNF(object):
 
         data: data to mask/unmask.
         """
+
+        if isinstance(mask_key, six.text_type):
+            mask_key = six.b(mask_key)
+
+        if isinstance(data, six.text_type):
+            data = six.b(data)
+
         _m = array.array("B", mask_key)
         _d = array.array("B", data)
-        for i in xrange(len(_d)):
+        for i in range(len(_d)):
             _d[i] ^= _m[i % 4]
-        return _d.tostring()
+
+        if six.PY3:
+            return _d.tobytes()
+        else:
+            return _d.tostring()
 
 
 class WebSocket(object):
@@ -391,11 +421,9 @@ class WebSocket(object):
         if sslopt is None:
             sslopt = {}
         self.connected = False
-        self.sock = socket.socket()
-        for opts in DEFAULT_SOCKET_OPTION:
-            self.sock.setsockopt(*opts)
-        for opts in sockopt:
-            self.sock.setsockopt(*opts)
+        self.sock = None
+        self._timeout = None
+        self.sockopt = sockopt
         self.sslopt = sslopt
         self.get_mask_key = get_mask_key
         self.fire_cont_frame = fire_cont_frame
@@ -427,7 +455,7 @@ class WebSocket(object):
         """
         Get the websocket timeout(second).
         """
-        return self.sock.gettimeout()
+        return self._timeout
 
     def settimeout(self, timeout):
         """
@@ -435,7 +463,7 @@ class WebSocket(object):
 
         timeout: timeout time(second).
         """
-        self.sock.settimeout(timeout)
+        self._timeout = timeout
 
     timeout = property(gettimeout, settimeout)
 
@@ -443,25 +471,49 @@ class WebSocket(object):
         """
         Connect to url. url is websocket url scheme. ie. ws://host:port/resource
         You can customize using 'options'.
-        If you set "header" dict object, you can set your own custom header.
+        If you set "header" list object, you can set your own custom header.
 
         >>> ws = WebSocket()
         >>> ws.connect("ws://echo.websocket.org/",
-                ...     header={"User-Agent: MyProgram",
-                ...             "x-custom: header"})
+                ...     header=["User-Agent: MyProgram",
+                ...             "x-custom: header"])
 
         timeout: socket timeout time. This value is integer.
                  if you set None for this value,
                  it means "use default_timeout value"
 
-        options: current support option is only "header".
-                 if you set header as dict value,
-                 the custom HTTP headers are added.
+        options: "header" -> custom http header list.
+                 "cookie" -> cookie value.
+                 "http_proxy_host" - http proxy host name.
+                 "http_proxy_port" - http proxy port. If not set, set to 80.
 
         """
+
         hostname, port, resource, is_secure = _parse_url(url)
+        proxy_host, proxy_port = options.get("http_proxy_host", None), options.get("http_proxy_port", 0)
+        if not proxy_host:
+            addrinfo_list = socket.getaddrinfo(hostname, port, 0, 0, socket.SOL_TCP)
+        else:
+            proxy_port = proxy_port and proxy_port or 80
+            addrinfo_list = socket.getaddrinfo(proxy_host, proxy_port, 0, 0, socket.SOL_TCP)
+
+        if not addrinfo_list:
+            raise WebSocketException("Host not found.: " + hostname + ":" + str(port))
+
+        family = addrinfo_list[0][0]
+        self.sock = socket.socket(family)
+        self.sock.settimeout(self.timeout)
+        for opts in DEFAULT_SOCKET_OPTION:
+            self.sock.setsockopt(*opts)
+        for opts in self.sockopt:
+            self.sock.setsockopt(*opts)
         # TODO: we need to support proxy
-        self.sock.connect((hostname, port))
+        address = addrinfo_list[0][4]
+        self.sock.connect(address)
+
+        if proxy_host:
+            self._tunnel(hostname, port)
+
         if is_secure:
             if HAVE_SSL:
                 sslopt = dict(cert_reqs=ssl.CERT_REQUIRED,
@@ -474,6 +526,21 @@ class WebSocket(object):
 
         self._handshake(hostname, port, resource, **options)
 
+    def _tunnel(self, host, port):
+        logger.debug("Connecting proxy...")
+        connect_header = "CONNECT %s:%d HTTP/1.0\r\n" % (host, port)
+        connect_header += "\r\n"
+        if traceEnabled:
+            logger.debug("--- request header ---")
+            logger.debug(connect_header)
+            logger.debug("-----------------------")
+
+        self._send(connect_header)
+
+        status, resp_headers = self._read_headers()
+        if status != 200:
+            raise WebSocketException("failed CONNECT via proxy")
+
     def _handshake(self, host, port, resource, **options):
         headers = []
         headers.append("GET %s HTTP/1.1" % resource)
@@ -493,8 +560,12 @@ class WebSocket(object):
         key = _create_sec_websocket_key()
         headers.append("Sec-WebSocket-Key: %s" % key)
         headers.append("Sec-WebSocket-Version: %s" % VERSION)
+        
         if "header" in options:
             headers.extend(options["header"])
+        cookie = options.get("cookie", None)
+        if cookie:
+          headers.append("Cookie: %s" % cookie)
 
         headers.append("")
         headers.append("")
@@ -519,7 +590,7 @@ class WebSocket(object):
         self.connected = True
 
     def _validate_header(self, headers, key):
-        for k, v in _HEADERS_TO_CHECK.iteritems():
+        for k, v in _HEADERS_TO_CHECK.items():
             r = headers.get(k, None)
             if not r:
                 return False
@@ -532,8 +603,11 @@ class WebSocket(object):
             return False
         result = result.lower()
 
-        value = key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
-        hashed = base64.encodestring(hashlib.sha1(value).digest()).strip().lower()
+        if isinstance(result, six.text_type):
+            result = result.encode('utf-8')
+
+        value = (key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11").encode('utf-8')
+        hashed = base64encode(hashlib.sha1(value).digest()).strip().lower()
         return hashed == result
 
     def _read_headers(self):
@@ -544,7 +618,8 @@ class WebSocket(object):
 
         while True:
             line = self._recv_line()
-            if line == "\r\n":
+            line = line.decode('utf-8')
+            if line == "\r\n" or line == "\n":
                 break
             line = line.strip()
             if traceEnabled:
@@ -575,6 +650,7 @@ class WebSocket(object):
 
         opcode: operation code to send. Please see OPCODE_XXX.
         """
+
         frame = ABNF.create_frame(payload, opcode)
         return self.send_frame(frame)
 
@@ -631,6 +707,8 @@ class WebSocket(object):
         return value: string(byte array) value.
         """
         opcode, data = self.recv_data()
+        if six.PY3 and opcode == ABNF.OPCODE_TEXT:
+            return data.decode("utf-8")
         return data
 
     def recv_data(self, control_frame=False):
@@ -715,17 +793,28 @@ class WebSocket(object):
 
         return value: ABNF frame object.
         """
+
         # Header
         if self._frame_header is None:
             self._frame_header = self._recv_strict(2)
-        b1 = ord(self._frame_header[0])
+
+        b1 = self._frame_header[0]
+
+        if six.PY2:
+            b1 = ord(b1)
+
         fin = b1 >> 7 & 1
         rsv1 = b1 >> 6 & 1
         rsv2 = b1 >> 5 & 1
         rsv3 = b1 >> 4 & 1
         opcode = b1 & 0xf
-        b2 = ord(self._frame_header[1])
+        b2 = self._frame_header[1]
+
+        if six.PY2:
+            b2 = ord(b2)
+
         has_mask = b2 >> 7 & 1
+
         # Frame length
         if self._frame_length is None:
             length_bits = b2 & 0x7f
@@ -737,17 +826,21 @@ class WebSocket(object):
                 self._frame_length = struct.unpack("!Q", length_data)[0]
             else:
                 self._frame_length = length_bits
+
         # Mask
         if self._frame_mask is None:
             self._frame_mask = self._recv_strict(4) if has_mask else ""
+
         # Payload
         payload = self._recv_strict(self._frame_length)
         if has_mask:
             payload = ABNF.mask(self._frame_mask, payload)
+
         # Reset for next frame
         self._frame_header = None
         self._frame_length = None
         self._frame_mask = None
+
         return ABNF(fin, rsv1, rsv2, rsv3, opcode, has_mask, payload)
 
 
@@ -798,13 +891,19 @@ class WebSocket(object):
         self.sock.close()
 
     def _send(self, data):
+
+        if isinstance(data, six.text_type):
+            data = data.encode('utf-8')
+
         try:
             return self.sock.send(data)
         except socket.timeout as e:
-            raise WebSocketTimeoutException(e.message)
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            raise WebSocketTimeoutException(message)
         except Exception as e:
-            if "timed out" in e.message:
-                raise WebSocketTimeoutException(e.message)
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            if "timed out" in message:
+                raise WebSocketTimeoutException(message)
             else:
                 raise
 
@@ -812,24 +911,28 @@ class WebSocket(object):
         try:
             bytes = self.sock.recv(bufsize)
         except socket.timeout as e:
-            raise WebSocketTimeoutException(e.message)
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            raise WebSocketTimeoutException(message)
         except SSLError as e:
-            if e.message == "The read operation timed out":
-                raise WebSocketTimeoutException(e.message)
+            message = getattr(e, 'strerror', getattr(e, 'message', ''))
+            if message == "The read operation timed out":
+                raise WebSocketTimeoutException(message)
             else:
                 raise
+
         if not bytes:
             raise WebSocketConnectionClosedException()
         return bytes
 
-
     def _recv_strict(self, bufsize):
         shortage = bufsize - sum(len(x) for x in self._recv_buffer)
         while shortage > 0:
             bytes = self._recv(shortage)
             self._recv_buffer.append(bytes)
             shortage -= len(bytes)
-        unified = "".join(self._recv_buffer)
+
+        unified = six.b("").join(self._recv_buffer)
+
         if shortage == 0:
             self._recv_buffer = []
             return unified
@@ -843,9 +946,9 @@ class WebSocket(object):
         while True:
             c = self._recv(1)
             line.append(c)
-            if c == "\n":
+            if c == six.b("\n"):
                 break
-        return "".join(line)
+        return six.b("").join(line)
 
 
 class WebSocketApp(object):
@@ -857,7 +960,7 @@ class WebSocketApp(object):
                  on_open=None, on_message=None, on_error=None,
                  on_close=None, on_ping=None, on_pong=None,
                  on_cont_message=None,
-                 keep_running=True, get_mask_key=None):
+                 keep_running=True, get_mask_key=None, cookie=None):
         """
         url: websocket url.
         header: custom header for websocket handshake.
@@ -885,6 +988,7 @@ class WebSocketApp(object):
         """
         self.url = url
         self.header = header
+        self.cookie = cookie
         self.on_open = on_open
         self.on_message = on_message
         self.on_error = on_error
@@ -903,6 +1007,7 @@ class WebSocketApp(object):
         data: message to send. If you set opcode to OPCODE_TEXT, data must be utf-8 string or unicode.
         opcode: operation code of data. default is OPCODE_TEXT.
         """
+
         if self.sock.send(data, opcode) == 0:
             raise WebSocketConnectionClosedException()
 
@@ -918,7 +1023,8 @@ class WebSocketApp(object):
             self.last_ping_tm = time.time()
             self.sock.ping()
 
-    def run_forever(self, sockopt=None, sslopt=None, ping_interval=0, ping_timeout=None):
+    def run_forever(self, sockopt=None, sslopt=None, ping_interval=0, ping_timeout=None,
+        http_proxy_host=None, http_proxy_port=None):
         """
         run event loop for WebSocket framework.
         This loop is infinite loop and is alive during websocket is available.
@@ -928,7 +1034,10 @@ class WebSocketApp(object):
         ping_interval: automatically send "ping" command every specified period(second)
             if set to 0, not send automatically.
         ping_timeout: timeout(second) if the pong message is not recieved.
+        http_proxy_host: http proxy host name.
+        http_proxy_port: http proxy port. If not set, set to 80.
         """
+
         if not ping_timeout or ping_timeout<=0:
             ping_timeout = None
         if sockopt is None:
@@ -943,7 +1052,8 @@ class WebSocketApp(object):
             self.sock = WebSocket(self.get_mask_key, sockopt=sockopt, sslopt=sslopt,
                 fire_cont_frame=self.on_cont_message and True or False)
             self.sock.settimeout(default_timeout)
-            self.sock.connect(self.url, header=self.header)
+            self.sock.connect(self.url, header=self.header, cookie=self.cookie,
+                http_proxy_host=http_proxy_host, http_proxy_port=http_proxy_port)
             self._callback(self.on_open)
 
             if ping_interval:
@@ -959,7 +1069,7 @@ class WebSocketApp(object):
                 if ping_timeout and self.last_ping_tm and time.time() - self.last_ping_tm > ping_timeout:
                     self.last_ping_tm = 0
                     raise WebSocketTimeoutException()
-                    
+
                 if r:
                     op_code, frame = self.sock.recv_data_frame(True)
                     if op_code == ABNF.OPCODE_CLOSE:
@@ -971,7 +1081,10 @@ class WebSocketApp(object):
                     elif op_code == ABNF.OPCODE_CONT and self.on_cont_message:
                         self._callback(self.on_cont_message, frame.data, frame.fin)
                     else:
-                        self._callback(self.on_message, frame.data)
+                        data = frame.data
+                        if six.PY3 and frame.opcode == ABNF.OPCODE_TEXT:
+                            data = data.decode("utf-8")
+                        self._callback(self.on_message, data)
         except Exception as e:
             self._callback(self.on_error, e)
         finally:
